/// <reference path="pal.Tizen.ts"/>
/// <reference path="main.ts"/>
/// <reference path="SourcePatcher.js"/>
/// <reference path="moment.min.js"/>
/// <reference path="semaphore.js" />

const baseDirectory: string = "downloads";
const cacheRetainDays: number = 7;

interface DownloadHTTPHeaderFields { };
enum DownloadNetworkType { "CELLULAR", "WIFI", "ALL" };

interface TizenDownloadRequest {
    url: string;
    destination?: string;
    fileName?: string;
    networkType?: DownloadNetworkType;
    httpHeader?: DownloadHTTPHeaderFields | any;
}

interface ResourceRequest {
    serverPath: string;
    serverUrl?: string;
    localFullDirectory?: string;
    localSubDirectory?: string
    localPath?: string;
    fileName?: string;
    lastModifiedDate?: string;
    downloadId?: number;
    dlTimeout?: number;
    tizenRequest?: TizenDownloadRequest;
}

const korbytStaticContent = [
    '/fonts/',
    '/images/',
    '/img/',
    '/jClocksGMT/',
    '/builder/1.5/'
];

class ResourceManager {
    private static downloadPossible: boolean;
    private static dlPending: Array<string> = [];
    private static dlDirectory;
    private static noNetworkLogCount: number;
    private static cleanCacheProcess: boolean;
    private static forceDownloads: boolean;

    private static finalCallback: (err) => any;

    private static manifestList: Array<string>;
    private static resourceKeeper: Array<string>;
    private static countFilesDownloaded: number;
    private static isDownloading: boolean = false;
    private static cacheCleanTimer;

    private static startTime: Date;
    private static finishTime: Date;

    private static fullDownloadPath: string;
    private static playerSplashFullPath: string;
    private static staticCache: any;

    private static new_etags: any;
    private static current_etags: any;
    private static maxConcurrentDownloads: number = 5;
    private static activeDownloads: string[] = [];

    static initializer() {
        ResourceManager.downloadPossible = tizen.systeminfo.getCapability('http://tizen.org/feature/download');
        _klogger.log(`downloadPossible: ${ResourceManager.downloadPossible}`);
        ResourceManager.dlPending = [];
        ResourceManager.staticCache = {};
        ResourceManager.isDownloading = false;
        ResourceManager.cacheCleanTimer = null;
        ResourceManager.new_etags = {};
        ResourceManager.current_etags = {};
        ResourceManager.FullDownloadPath = localStorage.getItem('fullDownloadPath');
        ResourceManager.maxConcurrentDownloads = 5;
        ResourceManager.activeDownloads = [];
    }

    static get FullDownloadPath(): string {
        return ResourceManager.fullDownloadPath;
    }

    static set FullDownloadPath(path: string) {
        if (ResourceManager.fullDownloadPath !== path) {
            ResourceManager.fullDownloadPath = path;
            SourcePatcher.baseUri = ResourceManager.fullDownloadPath;
            localStorage.setItem('fullDownloadPath', ResourceManager.fullDownloadPath);
        }
    }

    static get PlayerSplashFullPath(): string {
        return ResourceManager.playerSplashFullPath;
    }

    static set PlayerSplashFullPath(path: string) {
        if (ResourceManager.playerSplashFullPath !== path) {
            ResourceManager.playerSplashFullPath = path;
            localStorage.setItem('playerSplashFullPath', ResourceManager.playerSplashFullPath)
        }
    }

    private static isStaticContent(fn: string) {
        if (typeof fn === 'string') {
            for (let i = 0; i < korbytStaticContent.length; i++) {
                if (fn.startsWith(korbytStaticContent[i])) {
                    return true;
                }
            }
        }
        return false;
    }

    static getResources(cleanCache: boolean, resources: Array<string>, showProgress: boolean, callback: (err) => any) {
        ResourceManager.startTime = new Date();
        clearTimeout(ResourceManager.cacheCleanTimer);
        try {
            if (Array.isArray(resources) && resources.length > 0) {
                ResourceManager.isDownloading = true;
                ResourceManager.countFilesDownloaded = 0;

                // always include the manifest files in the keeper list.
                ResourceManager.resourceKeeper = ResourceManager.manifestList.slice();

                resources.forEach((file) => {
                    // append .HTML to Layouts, Messages, and Urgent Messages.
                    if (file.startsWith('/embed/') || file.startsWith('/message/') || file.startsWith('/urgent/')) {
                        file = file + '.html';
                    }

                    // Add each file to the keeper list
                    if (ResourceManager.resourceKeeper.indexOf(file) < 0) {
                        ResourceManager.resourceKeeper.push(file);
                    }
                });

                ResourceManager.resourceKeeper.push('/tizencached.html');
                ResourceManager.cleanCacheProcess = cleanCache;
                ResourceManager.dlPending = ResourceManager.dlPending.concat(resources.slice());

                // get a reference to the caching directory.
                return tizen.filesystem.resolve(baseDirectory, (dir) => {
                    ResourceManager.dlDirectory = dir;
                    if (ResourceManager.finalCallback != null) {
                        _klogger.error("ResourceManager.finalCallback is already set?");
                    }
                    ResourceManager.finalCallback = callback;
                    // return setTimeout(ResourceManager.downloadNext, 0);
                    if (showProgress) {
                        localStorage.setItem('KorbytDownloadProgress', 'starting');
                        PlayerHost.showResourceManagerProgress();
                    }
                    return ResourceManager.downloadNext();
                }, (err) => {
                    ResourceManager.isDownloading = false;
                    _klogger.error(err);
                    return callback(err);
                });
            }
        } catch (err) {
            _klogger.error(err);
            return callback(err);
        }
    }

    static deleteDownloadedFile(filePath: string, callback: () => any) {
        return tizen.filesystem.resolve(ResourceManager.getDirectory(filePath),
            (dir) => {
                dir.deleteFile(filePath, callback, callback);
            },
            (err) => {
                _klogger.debug(err);
                return callback();
            });
    }

    static start(tenantId: number, showProgress: boolean, callback: (err) => any) {
        try {
            clearTimeout(ResourceManager.cacheCleanTimer);
            ResourceManager.dlPending = [];
            ResourceManager.manifestList = [];
            if (ResourceManager.downloadPossible) {
                return ResourceManager.deleteDownloadedFile('downloads/prt/v3.8/manifest.json', () => {
                    ResourceManager.forceDownloads = true;
                    return tizen.filesystem.resolve('downloads/etags.json', function (etagsFile) {
                        try {
                            return etagsFile.readAsText((etags) => {
                                try {
                                    if (typeof etags === 'string') {
                                        etags = JSON.parse(etags);
                                    }
                                } catch (err) {
                                    console.log(etags);
                                    return ResourceManager.downloadManifest(tenantId, showProgress, function (err) {
                                        return callback(err);
                                    });
                                }
                                _klogger.debug(`ETAGS CONTENTS: ${JSON.stringify(etags)}`);
                                ResourceManager.current_etags = etags;
                                return ResourceManager.downloadManifest(tenantId, showProgress, function (err) {
                                    return callback(err);
                                });
                            });
                        } catch (err) {
                            _klogger.error(err);
                            console.log('ERROR Parsing ETAGS');
                            return ResourceManager.downloadManifest(tenantId, showProgress, function (err) {
                                return callback(err);
                            })
                        }
                    }, function (err) {
                        console.log('NO ETAGS.JSON');
                        return ResourceManager.downloadManifest(tenantId, showProgress, function (err) {
                            return callback(err);
                        })
                    });
                });
            }
        } catch (err) {
            _klogger.error(err);
            return callback(err);
        }
    }
    static downloadManifest(tenantId: number, showProgress: boolean, callback: (err) => any) {
        ResourceManager.forceDownloads = true;
        return ResourceManager.getResources(false, ['/prt/v3.8/manifest.json'], showProgress, () => {
            try {
                ResourceManager.forceDownloads = false;
                return tizen.filesystem.resolve('downloads/prt/v3.8/manifest.json',
                    (manifestFile) => {
                        try {
                            return manifestFile.readAsText(
                                (manifest) => {
                                    if (typeof manifest === 'string') {
                                        manifest = JSON.parse(manifest);
                                    }

                                    if (!Array.isArray(manifest.resources) || manifest.resources.length < 10) {
                                        throw new Error('player runtime manifest is invalid or empty');
                                    }

                                    _klogger.debug(`MANIFEST CONTENTS: ${JSON.stringify(manifest)}`);

                                    ResourceManager.resourceKeeper = [];
                                    let resources: Array<string> = [];
                                    if (!ResourceManager.playerSplashFullPath) {
                                        ResourceManager.playerSplashFullPath = `/ml/${tenantId}/branding/player_splash.jpg`;
                                    }
                                    resources.push(ResourceManager.playerSplashFullPath);

                                    for (let ndx = 0; ndx < manifest.resources.length; ndx++) {
                                        let fileName = manifest.resources[ndx].fn;
                                        resources.push(fileName);
                                    }
                                    ResourceManager.manifestList = resources;
                                    return ResourceManager.getResources(false, resources, showProgress, callback);
                                },
                                (err) => {
                                    _klogger.error(err);
                                    return callback(err);
                                });
                        } catch (err) {
                            _klogger.error(err);
                            return callback(err);
                        }
                    },
                    (err) => {
                        _klogger.error(err);
                        return callback(err);
                    });
            } catch (err) {
                _klogger.error(err);
                return callback(err);
            }
        });
    }
    static getFilename(path: string): string {
        try {
            return path.split('\\').pop().split('/').pop();
        } catch (err) {
            _klogger.error(err);
        }
    }

    static getDirectory(path: string): string {
        try {
            return path.substring(0, path.lastIndexOf("/"));
        } catch (err) {
            _klogger.error(err);
        }
    }

    static downloadCacheFile(dlResource: ResourceRequest) {
        try {
            if (!dlResource.serverUrl) {
                throw new Error('downloadCacheFile(): dlResource.serverUrl is not defined');
            }
            let url: string = dlResource.serverUrl + `?pid=${main.pal.playerID}`;
            let headers = { 'Cookie': `tx=${main.pal.provisioning.tenant_number}` };

            if (!ResourceManager.forceDownloads && dlResource.lastModifiedDate) {
                headers['If-Modified-Since'] = dlResource.lastModifiedDate;
            }

            if (!ResourceManager.forceDownloads && ResourceManager.current_etags[url]) {
                headers['If-None-Match'] = ResourceManager.current_etags[url];
            }

            $.ajax({
                type: "HEAD",
                url: url,
            }).done(function (message, text, jqXHR) {
                ResourceManager.new_etags[url] = jqXHR.getResponseHeader('etag');
            });
            dlResource.tizenRequest = new tizen.DownloadRequest(url, dlResource.localFullDirectory, dlResource.fileName, 'ALL', headers);
            dlResource.downloadId = tizen.download.start(dlResource.tizenRequest, {
                onprogress: (id, receivedSize, totalSize) => {
                    _klogger.debug(`##### ResourceManager.onProgress - ${id}: ${receivedSize} of ${totalSize}`);
                    clearTimeout(dlResource.dlTimeout);
                    dlResource.dlTimeout = setTimeout(ResourceManager.downloadOnTimeout, 60 * 1000);
                },
                onpaused: (id) => {
                    _klogger.info(`##### ResourceManager.onPaused - download paused: ${id}`);
                    clearTimeout(dlResource.dlTimeout);
                    dlResource.dlTimeout = setTimeout(ResourceManager.downloadOnTimeout, 60 * 1000);
                },
                oncanceled: (id) => {
                    _klogger.warn(`##### ResourceManager.onCanceled - download canceled: ${id}`);
                    dlResource.tizenRequest = null;
                    clearTimeout(dlResource.dlTimeout);

                    return ResourceManager.downloadNext();
                },
                oncompleted: (id, fullPath) => {
                    try {
                        _klogger.info(`##### ResourceManager.onCompleted - download completed: ${id} ${fullPath}`);
                        if (dlResource) {
                            clearTimeout(dlResource.dlTimeout);
                            dlResource.dlTimeout = null;

                            if (!ResourceManager.FullDownloadPath) {
                                ResourceManager.FullDownloadPath = fullPath.substr(0, fullPath.length - dlResource.serverPath.length);
                                SourcePatcher.baseUri = ResourceManager.FullDownloadPath;
                                _klogger.info(`~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ FullDownloadPath === ${ResourceManager.FullDownloadPath}`);
                            }

                            if (dlResource.serverPath.startsWith('/embed/') || dlResource.serverPath.startsWith('/message/') || dlResource.serverPath.startsWith('/urgent/')) {
                                dlResource.fileName = dlResource.fileName + '.html';
                            }
                            let orgName = ResourceManager.getFilename(dlResource.fileName);
                            let newName = ResourceManager.getFilename(fullPath);

                            if (ResourceManager.isStaticContent(dlResource.serverPath)) {
                                ResourceManager.staticCache[dlResource.serverPath] = true;
                            }

                            if (orgName !== newName) {
                                _klogger.debug(`~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ RENAME FILE ${fullPath} ===> ${newName} ===> ${orgName}`)
                                return ResourceManager.renameFile(fullPath, orgName, (fileName: string) => {
                                    return ResourceManager.postFileWork(fileName);
                                });
                            } else {
                                return ResourceManager.postFileWork(fullPath);
                            }
                        }
                    } catch (err) {
                        _klogger.error(err);
                    }
                },
                onfailed: (id, error) => {
                    dlResource.tizenRequest = null;
                    clearTimeout(dlResource.dlTimeout);

                    if(error.message && error.message.includes('HTTP status: 410')){
                        _klogger.error(`#### ResourceManager - Received 410 status code: ${id} => ${error}`);
                        main.deprovisioned();
                        main.setNextHeartbeat(true);
                        return;
                    }

                    if (error.code !== -1) {
                        _klogger.error(`##### ResourceManager - download failed: ${id} => ${error}`);
                    } else {
                        if (error.message && error.message.indexOf('HTTP status: 304') < 0) {
                            _klogger.debug(`##### ResourceManager - download failed: ${id} => ${error}`);
                        } else {
                            console.info(url + ": 304");
                            if (ResourceManager.isStaticContent(dlResource.serverPath)) {
                                ResourceManager.staticCache[dlResource.serverPath] = true;
                            }
                        }
                    }

                    return ResourceManager.downloadNext();
                }
            });

            dlResource.dlTimeout = setTimeout(() => {
                return ResourceManager.downloadOnTimeout(dlResource);
            }, 60 * 1000);

            _klogger.debug(`##### ResourceManager - started download #${dlResource.downloadId} of '${url}', remaining files ${ResourceManager.dlPending.length}`);
        } catch (err) {
            _klogger.error(err);
        }
    }

    static downloadOnTimeout(dlResource: ResourceRequest) {
        try {
            if (dlResource) {
                dlResource.tizenRequest = null;
                _klogger.info(`##### ResourceManager.onCompleted - DOWNLOAD TIMEOUT: ${dlResource.serverPath}`);
                tizen.download.cancel(dlResource.downloadId);
                clearTimeout(dlResource.dlTimeout);
            }
            return ResourceManager.downloadNext();
        } catch (err) {
            _klogger.error(err);
        }
    }

    static postFileWorkDone(err?: Error) {
        try {
            if (err) {
                _klogger.error(err);
            }

            // running count of files downloaded.
            ResourceManager.countFilesDownloaded++;

            // return setTimeout(ResourceManager.downloadNext, 0);
            return ResourceManager.downloadNext();
        } catch (err) {
            _klogger.error(err);
        }
    }

    static postFileWork(fullPath: string) {
        try {
            _klogger.debug(`##### ResourceManager.postFileWork - ${fullPath}`);
            let ext = fullPath.substr(fullPath.lastIndexOf('.') + 1).toLowerCase();
            if (ext === 'html' || ext === 'css' || ext === 'js') {
                if (ResourceManager.getFilename(fullPath).toLowerCase() != 'playerengine.js') {
                    return tizen.filesystem.resolve(fullPath,
                        (file) => {
                            return file.readAsText(
                                (fileText) => {
                                    try {
                                        let patchedText = SourcePatcher.patch(fileText);
                                        // if (ext === 'css') {
                                        //     _klogger.debug(patchedText);
                                        // }
                                        if (patchedText !== fileText) {
                                            _klogger.info(`##### ResourceManager - write patched file ${file.fullPath}`);
                                            return file.openStream("w",
                                                (fStream) => {
                                                    fStream.write(patchedText);
                                                    fStream.close();
                                                    return ResourceManager.postFileWorkDone();
                                                },
                                                (err) => {
                                                    return ResourceManager.postFileWorkDone(err);
                                                }, "UTF-8");
                                        } else {
                                            return ResourceManager.postFileWorkDone();
                                        }
                                    } catch (err) {
                                        _klogger.error(err);
                                    }
                                },
                                (err) => {
                                    return ResourceManager.postFileWorkDone(err);
                                });
                        },
                        (err) => {
                            return ResourceManager.postFileWorkDone(err);
                        });
                }
            }
        } catch (err) {
            return ResourceManager.postFileWorkDone(err);
        }
        return ResourceManager.postFileWorkDone();
    }

    static renameFile(sourcePath: string, fileName: string, callback: (fullPath: string) => any) {
        try {
            let srcDirectory = ResourceManager.getDirectory(sourcePath);
            let fullPath = srcDirectory + '/' + fileName;
            return tizen.filesystem.resolve(srcDirectory, (srcFile) => {
                return srcFile.moveTo(sourcePath, fullPath, true,
                    () => {
                        _klogger.info(`##### ResourceManager - FILE MOVED ${sourcePath} ==> ${fullPath}`);
                        return callback(fullPath);
                    },
                    (err) => {
                        _klogger.error(`##### ResourceManager - FILE MOVED FAILED ${sourcePath} ==> ${fullPath}\n${err}`);
                        return callback(fullPath);
                    });
            }, (err) => {
                _klogger.error(err);
                return callback(sourcePath);
            });
        } catch (err) {
            _klogger.error(err);
        }
    }

    static getFileInfo(dlResource: ResourceRequest) {
        try {
            tizen.filesystem.resolve(dlResource.localPath, (newFile) => {
                dlResource.lastModifiedDate = newFile.modified.toUTCString();
                ResourceManager.downloadCacheFile(dlResource);
            }, (err) => {
                if (err.code === 8) {
                    return ResourceManager.downloadCacheFile(dlResource);
                }

                _klogger.error(err);
            });
        } catch (err) {
            _klogger.error(err);
            // return setTimeout(ResourceManager.downloadNext, 0);
            return ResourceManager.downloadNext();
        }
    }

    static createCacheDirectory(dlResource: ResourceRequest) {
        _klogger.debug(`##### ResourceManager - createCacheDirectory(${JSON.stringify(dlResource)})`);
        try {
            ResourceManager.dlDirectory.createDirectory(dlResource.localSubDirectory);
            ResourceManager.downloadCacheFile(dlResource);
        } catch (err) {
            _klogger.error(err);
            // return setTimeout(ResourceManager.downloadNext, 0);
            return ResourceManager.downloadNext();
        }
    }

    static checkCacheDirectory(dlResource: ResourceRequest) {
        // _klogger.debug(`##### ResourceManager - checkCacheDirectory(${JSON.stringify(dlResource)})`);
        try {
            tizen.filesystem.resolve(dlResource.localFullDirectory,
                (newPath) => {
                    return ResourceManager.getFileInfo(dlResource);
                }, (err) => {
                    try {
                        if (err.code === 8) {
                            return ResourceManager.createCacheDirectory(dlResource);
                        }
                        _klogger.error(err);
                    } catch (err) {
                        _klogger.error(err);
                    }

                    return ResourceManager.downloadNext();
                });
        } catch (err) {
            _klogger.error(err);
            return ResourceManager.downloadNext();
        }
    }

    static downloadNext() {
        try {
            if (!ResourceManager.dlPending || ResourceManager.dlPending.length < 1) {
                return ResourceManager.downloadComplete();
            }

            return tizen.systeminfo.getPropertyValue('NETWORK', function (networkInfo) {
                try {
                    if (networkInfo.networkType === 'NONE') {
                        if (ResourceManager.noNetworkLogCount <= 0) {
                            ResourceManager.noNetworkLogCount = 30;
                            _klogger.warn('##### ResourceManager - Network connection is not available. Download is not possible.');
                        } else {
                            ResourceManager.noNetworkLogCount--;
                        }
                        return setTimeout(ResourceManager.downloadNext, 10 * 1000);
                    }

                    if (ResourceManager.dlPending.length > 0 && ResourceManager.activeDownloads.length < ResourceManager.maxConcurrentDownloads) {
                        let remaining = ResourceManager.resourceKeeper.length - ResourceManager.dlPending.length;
                        let percent = (((remaining) / ResourceManager.resourceKeeper.length) * 100).toFixed(2);
                        localStorage.setItem('KorbytDownloadProgress', `${remaining}/${ResourceManager.resourceKeeper.length}(${percent}%)`);

                        let dlResource: ResourceRequest = {
                            serverPath: ResourceManager.dlPending.shift(),
                        };

                        if (ResourceManager.isStaticContent(dlResource.serverPath)) {
                            if (ResourceManager.staticCache[dlResource.serverPath]) {
                                // _klogger.info(`##### ResourceManager - Already downloaded: ${dlResource.serverPath}`);
                                return ResourceManager.downloadNext(); // Skip and go onto next item
                            }
                        }

                        dlResource.serverUrl = main.pal.provisioning.cms + dlResource.serverPath;
                        if (dlResource.serverPath == '/prt/v3.8/tizencached.html') {
                            dlResource.localPath = ResourceManager.dlDirectory.path + '/' + ResourceManager.getFilename(dlResource.serverPath);
                        } else {
                            dlResource.localPath = ResourceManager.dlDirectory.path + dlResource.serverPath;
                        }

                        dlResource.localFullDirectory = ResourceManager.getDirectory(dlResource.localPath);
                        dlResource.localSubDirectory = ResourceManager.getDirectory(dlResource.serverPath);
                        dlResource.fileName = ResourceManager.getFilename(dlResource.localPath);
                        ResourceManager.activeDownloads.push(dlResource.serverPath);

                        // start the download process...
                        return ResourceManager.checkCacheDirectory(dlResource);
                    } else {
                        return ResourceManager.downloadComplete();
                    }
                } catch (err) {
                    _klogger.error(`##### ResourceManager - ${err.stack || err.message}`);
                }
            });
        } catch (err) {
            _klogger.error(`##### ResourceManager - ${err.stack || err.message}`);
        }

        // return setTimeout(ResourceManager.downloadNext, 0);
        return ResourceManager.downloadNext();
    }

    static cleanerDirectoryList: Array<any>;
    static cleanerFilesList: Array<any>;
    static cacheDeleteCutoff;
    static cacheCleanDirectory(wipeYourCache: boolean, callback: () => any) {
        try {
            if (!ResourceManager.cleanerDirectoryList || ResourceManager.cleanerDirectoryList.length < 1) {
                return ResourceManager.cacheDeleteFilesList(wipeYourCache, () => {
                    _klogger.debug(`##### ResourceManager - completed cleanup of file cache`);
                    return callback();
                });
            }

            let dir = ResourceManager.cleanerDirectoryList.pop();
            return dir.listFiles((files) => {
                try {
                    for (let ndx = 0; ndx < files.length; ndx++) {
                        let file = files[ndx];
                        if (file.isDirectory) {
                            // _klogger.debug(`ResourceManager.cleanerDirectoryList.push(${file.fullPath})`);
                            ResourceManager.cleanerDirectoryList.push(file);
                        } else {
                            // _klogger.debug(`ResourceManager.cleanerFilesList.push(${file.fullPath})`);
                            ResourceManager.cleanerFilesList.push(file);
                        }
                    }

                    ResourceManager.cacheCleanDirectory(wipeYourCache, callback);
                } catch (err) {
                    _klogger.error(err);
                    return callback();
                }
            }, (err) => {
                _klogger.error(err);
                return callback();
            });
        } catch (err) {
            _klogger.error(err);
            return callback();
        }
    }

    static cacheDeleteFilesList(wipeYourCache: boolean, callback: () => any) {
        try {
            if (!ResourceManager.cleanerFilesList || ResourceManager.cleanerFilesList.length < 1) {
                return callback();
            }

            let file = ResourceManager.cleanerFilesList.pop();
            let path = file.fullPath.substring(baseDirectory.length);

            if (wipeYourCache || (ResourceManager.resourceKeeper.indexOf(path) < 0 && !file.fullPath.startsWith('downloads/prt/'))) {
                //if (wipeYourCache || ResourceManager.cacheDeleteCutoff.isAfter(file.modified)) 
                {
                    return ResourceManager.deleteDownloadedFile(file.fullPath, () => {
                        _klogger.debug(`##### ResourceManager ----------------------------------------------------------------------- DELETE ${file.fullPath}`);
                        return ResourceManager.cacheDeleteFilesList(wipeYourCache, callback);
                    });
                }
            }

            return setTimeout(() => {
                return ResourceManager.cacheDeleteFilesList(wipeYourCache, callback);
            }, 0);
        } catch (err) {
            _klogger.error(err);
            return callback();
        }
    }

    static cacheCleaner(wipeYourCache?: boolean, callback?: () => any) {
        try {
            if (!callback) {
                callback = () => { }
            }
            ResourceManager.cleanerDirectoryList = [];
            ResourceManager.cleanerFilesList = [];
            clearTimeout(ResourceManager.cacheCleanTimer);
            // currently downloading?
            if (wipeYourCache || (ResourceManager.resourceKeeper.length > 0 && !ResourceManager.isDownloading)) {
                if (wipeYourCache) {
                    ResourceManager.resourceKeeper = [];
                }
                _klogger.info('##### ResourceManager - cache clean up starting NOW');
                ResourceManager.cacheDeleteCutoff = moment().add(-cacheRetainDays, 'days');
                return tizen.filesystem.resolve(baseDirectory, (downloads) => {
                    ResourceManager.cleanerDirectoryList.push(downloads);
                    ResourceManager.cacheCleanDirectory(wipeYourCache, callback);
                }, (err) => {
                    _klogger.error(err);
                    return callback();
                });
            } else {
                ResourceManager.startCacheCleaner();
                return callback();
            }
        } catch (err) {
            _klogger.error(err);
        }
    }

    private static startCacheCleaner() {
        if (ResourceManager.countFilesDownloaded < 1) {
            _klogger.debug(`##### ResourceManager - no files downloaded, skipping cache clean up pass.`);
        } else {
            _klogger.debug(`##### ResourceManager - starting wait timer for cache clean up`);
            clearTimeout(ResourceManager.cacheCleanTimer);
            ResourceManager.cacheCleanTimer = setTimeout(ResourceManager.cacheCleaner, 30 * 1000);
        }
    }

    private static downloadComplete() {
        try {
            // if the company splash file exists update localStorage of launch page can use it.
            if (ResourceManager.playerSplashFullPath && ResourceManager.FullDownloadPath) {
                ResourceManager.playerSplashFullPath = SourcePatcher.fixPath(ResourceManager.playerSplashFullPath);
                tizen.filesystem.resolve(ResourceManager.playerSplashFullPath,
                    (splashFile) => {
                        if (splashFile.isFile) {
                            localStorage.setItem('KorbytPlayerSplashPath', ResourceManager.playerSplashFullPath);
                        }
                    },
                    (err) => {
                        if (err.code === 8) {
                            _klogger.info(`player splash graphic is not installed`);
                        } else {
                            _klogger.error(err);
                        }
                    });
            }

            localStorage.removeItem('KorbytDownloadProgress');

            if (!ResourceManager.isDownloading) {
                _klogger.error(`ResourceManager.downloadComplete() - ResourceManager.isDownloading === ${ResourceManager.isDownloading}`);
            }

            ResourceManager.isDownloading = false;

            ResourceManager.finishTime = new Date();
            let downloadTime = ResourceManager.finishTime.getTime() - ResourceManager.startTime.getTime();
            _klogger.info(`##### ResourceManager - DOWNLOAD COMPLETED, downloaded ${ResourceManager.countFilesDownloaded} of ${ResourceManager.resourceKeeper.length} files in ${downloadTime / 1000} seconds`);
            if (ResourceManager.cleanCacheProcess) {
                ResourceManager.startCacheCleaner();
            }

            if (ResourceManager.new_etags) {
                return ResourceManager.deleteDownloadedFile('downloads/etags.json', function () {
                    tizen.filesystem.resolve('downloads',
                        (dir) => {
                            var etags_file = dir.createFile("etags.json");
                            if (etags_file != null) {
                                return etags_file.openStream('rw', function (fs) {
                                    try {
                                        fs.write(JSON.stringify(ResourceManager.new_etags));
                                        fs.close();
                                        if (ResourceManager.finalCallback) {
                                            ResourceManager.finalCallback(null);
                                            ResourceManager.finalCallback = null;
                                        } else {
                                            _klogger.warn('ResourceManager.downloadComplete() - no final callback method found.');
                                        }
                                    } catch (err) {
                                        _klogger.error('ResourceManager.downloadComplete().etags_file.openStream() Exception - ' + err);
                                    }
                                }, null, 'UTF-8');
                            }
                        },
                        (err) => {
                            _klogger.error('Failed resolve directory "downloads": ' + err);
                            if (ResourceManager.finalCallback) {
                                ResourceManager.finalCallback(null);
                                ResourceManager.finalCallback = null;
                            } else {
                                _klogger.warn('ResourceManager.downloadComplete() - no final callback method found.');
                            }
                        });
                });
            }
            if (ResourceManager.finalCallback) {
                ResourceManager.finalCallback(null);
                ResourceManager.finalCallback = null;
            } else {
                _klogger.warn('ResourceManager.downloadComplete() - no final callback method found.');
            }
        } catch (err) {
            _klogger.error(err);
        }
    }
}